@@ -1,20 +1,24 @@
 /*
- * Copyright (C) 2012-2017 sysmocom - s.f.m.c. GmbH
+ * Copyright (C) 2018-2019 sysmocom - s.f.m.c. GmbH
  * All Rights Reserved
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * SPDX-License-Identifier: AGPL-3.0+
+ *
+ * Author: Pau Espin Pedrol <pespin@sysmocom.de>
+ *
+ * This program is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU Affero General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Affero General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
+ * You should have received a copy of the GNU Affero General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
+ * See the COPYING file in the main directory for details.
  */
 
 #include <string.h>
@@ -31,6 +35,7 @@
 #include <osmocom/vty/vty.h>
 #include <osmocom/vty/misc.h>
 
+#include "trx_rate_ctr.h"
 #include "trx_vty.h"
 #include "../config.h"
 
@@ -43,15 +48,25 @@ static const struct value_string clock_ref_names[] = {
 	{ 0,		NULL }
 };
 
-static const struct value_string filler_names[] = {
-	{ FILLER_DUMMY,		"Dummy bursts" },
-	{ FILLER_ZERO,		"Disabled" },
-	{ FILLER_NORM_RAND,	"Normal bursts with random payload" },
-	{ FILLER_EDGE_RAND,	"EDGE bursts with random payload" },
-	{ FILLER_ACCESS_RAND,	"Access bursts with random payload" },
+const struct value_string filler_names[] = {
+	{ FILLER_DUMMY,		"Dummy bursts (C0 only)" },
+	{ FILLER_ZERO,		"Empty bursts" },
+	{ FILLER_NORM_RAND,	"GMSK Normal Bursts with random payload" },
+	{ FILLER_EDGE_RAND,	"8-PSK Normal Bursts with random payload" },
+	{ FILLER_ACCESS_RAND,	"Access Bursts with random payload" },
 	{ 0,			NULL }
 };
 
+static const struct value_string filler_types[] = {
+	{ FILLER_DUMMY,		"dummy" },
+	{ FILLER_ZERO,		"zero" },
+	{ FILLER_NORM_RAND,	"random-nb-gmsk" },
+	{ FILLER_EDGE_RAND,	"random-nb-8psk" },
+	{ FILLER_ACCESS_RAND,	"random-ab" },
+	{ 0,			NULL }
+};
+
+
 struct trx_ctx *trx_from_vty(struct vty *v)
 {
         /* It can't hurt to force callers to continue to pass the vty instance
@@ -167,57 +182,10 @@ DEFUN(cfg_rx_sps, cfg_rx_sps_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_test_rtsc, cfg_test_rtsc_cmd,
-	"test rtsc <0-7>",
-	"Set the Random Normal Burst test mode with TSC\n"
-	"TSC\n")
-{
-	struct trx_ctx *trx = trx_from_vty(vty);
-
-	if (trx->cfg.rach_delay_set) {
-		vty_out(vty, "rach-delay and rtsc options are mutual-exclusive%s",
-			VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	trx->cfg.rtsc_set = true;
-	trx->cfg.rtsc = atoi(argv[0]);
-	if (!trx->cfg.egprs) /* Don't override egprs which sets different filler */
-		trx->cfg.filler = FILLER_NORM_RAND;
-
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_test_rach_delay, cfg_test_rach_delay_cmd,
-	"test rach-delay <0-68>",
-	"Set the Random Access Burst test mode with delay\n"
-	"RACH delay\n")
-{
-	struct trx_ctx *trx = trx_from_vty(vty);
-
-	if (trx->cfg.rtsc_set) {
-		vty_out(vty, "rach-delay and rtsc options are mutual-exclusive%s",
-			VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	if (trx->cfg.egprs) {
-		vty_out(vty, "rach-delay and egprs options are mutual-exclusive%s",
-			VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	trx->cfg.rach_delay_set = true;
-	trx->cfg.rach_delay = atoi(argv[0]);
-	trx->cfg.filler = FILLER_ACCESS_RAND;
-
-	return CMD_SUCCESS;
-}
-
 DEFUN(cfg_clock_ref, cfg_clock_ref_cmd,
 	"clock-ref (internal|external|gpsdo)",
 	"Set the Reference Clock\n"
-	"Enable internal referece (default)\n"
+	"Enable internal reference (default)\n"
 	"Enable external 10 MHz reference\n"
 	"Enable GPSDO reference\n")
 {
@@ -300,7 +268,6 @@ DEFUN(cfg_egprs, cfg_egprs_cmd,
 		trx->cfg.egprs = false;
 	} else if (strcmp("enable", argv[0]) == 0) {
 		trx->cfg.egprs = true;
-		trx->cfg.filler = FILLER_EDGE_RAND;
 	} else {
 		return CMD_WARNING;
 	}
@@ -335,14 +302,171 @@ DEFUN(cfg_rt_prio, cfg_rt_prio_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_filler, cfg_filler_cmd,
-	"filler dummy",
-	"Enable C0 filler table\n"
-	"Dummy method\n")
+DEFUN(cfg_stack_size, cfg_stack_size_cmd,
+	"stack-size <0-2147483647>",
+	"Set the stack size per thread in BYTE, 0 = OS default\n"
+	"Stack size per thread in BYTE\n")
 {
 	struct trx_ctx *trx = trx_from_vty(vty);
 
-	trx->cfg.filler = FILLER_DUMMY;
+	trx->cfg.stack_size = atoi(argv[0]);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_filler, cfg_filler_type_cmd,
+	"filler type (zero|dummy|random-nb-gmsk|random-nb-8psk|random-ab)",
+	"Filler burst settings\n"
+	"Filler burst type (default=zero)\n"
+	"Send an empty burst when there is nothing to send (default)\n"
+	"Send a dummy burst when there is nothing to send on C0 (TRX0) and empty burst on other channels."
+	" Use for OpenBTS compatibility only, don't use with OsmoBTS as it breaks encryption.\n"
+	"Send a GMSK modulated Normal Burst with random bits when there is nothing to send."
+	" Use for spectrum mask testing. Configure 'filler tsc' to set training sequence.\n"
+	"Send an 8-PSK modulated Normal Burst with random bits when there is nothing to send."
+	" Use for spectrum mask testing. Configure 'filler tsc' to set training sequence.\n"
+	"Send an Access Burst with random bits when there is nothing to send. Use for Rx/Tx alignment."
+	" Configure 'filler access-burst-delay' to introduce artificial delay.\n"
+)
+{
+	struct trx_ctx *trx = trx_from_vty(vty);
+	// trx->cfg.filler is unsigned, so we need an interim int var to detect errors
+	int type = get_string_value(filler_types, argv[0]);
+
+	if (type < 0) {
+		trx->cfg.filler = FILLER_ZERO;
+		return CMD_WARNING;
+	}
+	trx->cfg.filler = type;
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_test_rtsc, cfg_filler_tsc_cmd,
+	"filler tsc <0-7>",
+	"Filler burst settings\n"
+	"Set the TSC for GMSK/8-PSK Normal Burst random fillers. Used only with 'random-nb-gmsk' and"
+	" 'random-nb-8psk' filler types. (default=0)\n"
+	"TSC\n")
+{
+	struct trx_ctx *trx = trx_from_vty(vty);
+
+	trx->cfg.rtsc = atoi(argv[0]);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_test_rach_delay, cfg_filler_rach_delay_cmd,
+	"filler access-burst-delay <0-68>",
+	"Filler burst settings\n"
+	"Set the delay for Access Burst random fillers. Used only with 'random-ab' filler type. (default=0)\n"
+	"RACH delay in symbols\n")
+{
+	struct trx_ctx *trx = trx_from_vty(vty);
+
+	trx->cfg.rach_delay = atoi(argv[0]);
+
+	return CMD_SUCCESS;
+}
+
+static int vty_ctr_name_2_id(const char* str) {
+	size_t i;
+	for (i = 0; trx_chan_ctr_names[i].str; i++) {
+		if (strstr(trx_chan_ctr_names[i].str, str)) {
+			return i;
+		}
+	}
+	return -1;
+}
+
+static int vty_intv_name_2_id(const char* str) {
+	size_t i;
+	for (i = 0; rate_ctr_intv[i].str; i++) {
+		if (strcmp(rate_ctr_intv[i].str, str) == 0) {
+			return i;
+		}
+	}
+	return -1;
+}
+
+#define THRESHOLD_ARGS "(rx_overruns|tx_underruns|rx_drop_events|rx_drop_samples|tx_drop_events|tx_drop_samples)"
+#define THRESHOLD_STR_VAL(s) "Set threshold value for rate_ctr device:" OSMO_STRINGIFY_VAL(s) "\n"
+#define THRESHOLD_STRS \
+	THRESHOLD_STR_VAL(rx_overruns) \
+	THRESHOLD_STR_VAL(tx_underruns) \
+	THRESHOLD_STR_VAL(rx_drop_events) \
+	THRESHOLD_STR_VAL(rx_drop_samples) \
+	THRESHOLD_STR_VAL(tx_drop_events) \
+	THRESHOLD_STR_VAL(tx_drop_samples)
+#define INTV_ARGS "(per-second|per-minute|per-hour|per-day)"
+#define INTV_STR_VAL(s) "Threshold value sampled " OSMO_STRINGIFY_VAL(s) "\n"
+#define INTV_STRS \
+	INTV_STR_VAL(per-second) \
+	INTV_STR_VAL(per-minute) \
+	INTV_STR_VAL(per-hour) \
+	INTV_STR_VAL(per-day)
+
+DEFUN(cfg_ctr_error_threshold, cfg_ctr_error_threshold_cmd,
+	"ctr-error-threshold " THRESHOLD_ARGS " <0-65535> " INTV_ARGS,
+	"Threshold rate for error counter\n"
+	THRESHOLD_STRS
+	"Value to set for threshold\n"
+	INTV_STRS)
+{
+	int rc;
+	struct ctr_threshold ctr;
+
+	struct trx_ctx *trx = trx_from_vty(vty);
+	rc = vty_ctr_name_2_id(argv[0]);
+	if (rc < 0) {
+		vty_out(vty, "No valid ctr_name found for ctr-error-threshold %s%s",
+			argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	ctr.ctr_id = (enum TrxCtr)rc;
+	ctr.val = atoi(argv[1]);
+	rc = vty_intv_name_2_id(argv[2]);
+	if (rc < 0) {
+		vty_out(vty, "No valid time frame found for ctr-error-threshold %s %d %s%s",
+			argv[0], ctr.val, argv[2], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	ctr.intv = (enum rate_ctr_intv) rc;
+	trx_rate_ctr_threshold_add(&ctr);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_no_ctr_error_threshold, cfg_no_ctr_error_threshold_cmd,
+	"no ctr-error-threshold " THRESHOLD_ARGS " <0-65535> " INTV_ARGS,
+	NO_STR "Threshold rate for error counter\n"
+	THRESHOLD_STRS
+	"Value to set for threshold\n"
+	INTV_STRS)
+{
+	int rc;
+	struct ctr_threshold ctr;
+
+	struct trx_ctx *trx = trx_from_vty(vty);
+	rc = vty_ctr_name_2_id(argv[0]);
+	if (rc < 0) {
+		vty_out(vty, "No valid ctr_name found for ctr-error-threshold %s%s",
+			argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	ctr.ctr_id = (enum TrxCtr)rc;
+	ctr.val = atoi(argv[1]);
+	rc = vty_intv_name_2_id(argv[2]);
+	if (rc < 0) {
+		vty_out(vty, "No valid time frame found for ctr-error-threshold %s %d %s%s",
+			argv[0], ctr.val, argv[2], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	ctr.intv = (enum rate_ctr_intv) rc;
+	if (trx_rate_ctr_threshold_del(&ctr) < 0) {
+		vty_out(vty, "no ctr-error-threshold: Entry to delete not found%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	}
 
 	return CMD_SUCCESS;
 }
@@ -428,10 +552,6 @@ static int config_write_trx(struct vty *vty)
 		vty_out(vty, " tx-sps %u%s", trx->cfg.tx_sps, VTY_NEWLINE);
 	if (trx->cfg.rx_sps != DEFAULT_RX_SPS)
 		vty_out(vty, " rx-sps %u%s", trx->cfg.rx_sps, VTY_NEWLINE);
-	if (trx->cfg.rtsc_set)
-		vty_out(vty, " test rtsc %u%s", trx->cfg.rtsc, VTY_NEWLINE);
-	if (trx->cfg.rach_delay_set)
-		vty_out(vty, " test rach-delay %u%s", trx->cfg.rach_delay, VTY_NEWLINE);
 	if (trx->cfg.clock_ref != REF_INTERNAL)
 		vty_out(vty, " clock-ref %s%s", get_value_string(clock_ref_names, trx->cfg.clock_ref), VTY_NEWLINE);
 	vty_out(vty, " multi-arfcn %s%s", trx->cfg.multi_arfcn ? "enable" : "disable", VTY_NEWLINE);
@@ -444,6 +564,15 @@ static int config_write_trx(struct vty *vty)
 	vty_out(vty, " ext-rach %s%s", trx->cfg.ext_rach ? "enable" : "disable", VTY_NEWLINE);
 	if (trx->cfg.sched_rr != 0)
 		vty_out(vty, " rt-prio %u%s", trx->cfg.sched_rr, VTY_NEWLINE);
+	if (trx->cfg.filler != FILLER_ZERO)
+		vty_out(vty, " filler type %s%s", get_value_string(filler_types, trx->cfg.filler), VTY_NEWLINE);
+	if (trx->cfg.rtsc > 0)
+		vty_out(vty, " filler tsc %u%s", trx->cfg.rtsc, VTY_NEWLINE);
+	if (trx->cfg.rach_delay > 0)
+		vty_out(vty, " filler access-burst-delay %u%s", trx->cfg.rach_delay, VTY_NEWLINE);
+	if (trx->cfg.stack_size != 0)
+		vty_out(vty, " stack-size %u%s", trx->cfg.stack_size, VTY_NEWLINE);
+	trx_rate_ctr_threshold_write_config(vty, " ");
 
 	for (i = 0; i < trx->cfg.num_chans; i++) {
 		chan = &trx->cfg.chans[i];
@@ -468,11 +597,9 @@ static void trx_dump_vty(struct vty *vty, struct trx_ctx *trx)
 	vty_out(vty, " Device args: %s%s", trx->cfg.dev_args, VTY_NEWLINE);
 	vty_out(vty, " Tx Samples-per-Symbol: %u%s", trx->cfg.tx_sps, VTY_NEWLINE);
 	vty_out(vty, " Rx Samples-per-Symbol: %u%s", trx->cfg.rx_sps, VTY_NEWLINE);
-	vty_out(vty, " Test Mode: TSC: %u (%s)%s", trx->cfg.rtsc,
-		trx->cfg.rtsc_set ? "Enabled" : "Disabled", VTY_NEWLINE);
-	vty_out(vty, " Test Mode: RACH Delay: %u (%s)%s", trx->cfg.rach_delay,
-		trx->cfg.rach_delay_set ? "Enabled" : "Disabled", VTY_NEWLINE);
-	vty_out(vty, " C0 Filler Table: %s%s", get_value_string(filler_names, trx->cfg.filler), VTY_NEWLINE);
+	vty_out(vty, " Filler Burst Type: %s%s", get_value_string(filler_names, trx->cfg.filler), VTY_NEWLINE);
+	vty_out(vty, " Filler Burst TSC: %u%s", trx->cfg.rtsc, VTY_NEWLINE);
+	vty_out(vty, " Filler Burst RACH Delay: %u%s", trx->cfg.rach_delay, VTY_NEWLINE);
 	vty_out(vty, " Clock Reference: %s%s", get_value_string(clock_ref_names, trx->cfg.clock_ref), VTY_NEWLINE);
 	vty_out(vty, " Multi-Carrier: %s%s", trx->cfg.multi_arfcn ? "Enabled" : "Disabled", VTY_NEWLINE);
 	vty_out(vty, " Tuning offset: %f%s", trx->cfg.offset, VTY_NEWLINE);
@@ -482,6 +609,7 @@ static void trx_dump_vty(struct vty *vty, struct trx_ctx *trx)
 	vty_out(vty, " Extended RACH support: %s%s", trx->cfg.ext_rach ? "Enabled" : "Disabled", VTY_NEWLINE);
 	vty_out(vty, " Real Time Priority: %u (%s)%s", trx->cfg.sched_rr,
 		trx->cfg.sched_rr ? "Enabled" : "Disabled", VTY_NEWLINE);
+	vty_out(vty, " Stack size per Thread in BYTE (0 = OS default): %u%s", trx->cfg.stack_size, VTY_NEWLINE);
 	vty_out(vty, " Channels: %u%s", trx->cfg.num_chans, VTY_NEWLINE);
 	for (i = 0; i < trx->cfg.num_chans; i++) {
 		chan = &trx->cfg.chans[i];
@@ -540,6 +668,7 @@ static int trx_vty_go_parent(struct vty *vty)
 static const char trx_copyright[] =
 	"Copyright (C) 2007-2014 Free Software Foundation, Inc.\r\n"
 	"Copyright (C) 2013 Thomas Tsou <tom@tsou.cc>\r\n"
+	"Copyright (C) 2013-2019 Fairwaves, Inc.\r\n"
 	"Copyright (C) 2015 Ettus Research LLC\r\n"
 	"Copyright (C) 2017-2018 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>\r\n"
 	"License AGPLv3+: GNU AGPL version 3 or later <http://gnu.org/licenses/agpl-3.0.html>\r\n"
@@ -582,8 +711,6 @@ int trx_vty_init(struct trx_ctx* trx)
 	install_element(TRX_NODE, &cfg_dev_args_cmd);
 	install_element(TRX_NODE, &cfg_tx_sps_cmd);
 	install_element(TRX_NODE, &cfg_rx_sps_cmd);
-	install_element(TRX_NODE, &cfg_test_rtsc_cmd);
-	install_element(TRX_NODE, &cfg_test_rach_delay_cmd);
 	install_element(TRX_NODE, &cfg_clock_ref_cmd);
 	install_element(TRX_NODE, &cfg_multi_arfcn_cmd);
 	install_element(TRX_NODE, &cfg_offset_cmd);
@@ -592,7 +719,12 @@ int trx_vty_init(struct trx_ctx* trx)
 	install_element(TRX_NODE, &cfg_egprs_cmd);
 	install_element(TRX_NODE, &cfg_ext_rach_cmd);
 	install_element(TRX_NODE, &cfg_rt_prio_cmd);
-	install_element(TRX_NODE, &cfg_filler_cmd);
+	install_element(TRX_NODE, &cfg_filler_type_cmd);
+	install_element(TRX_NODE, &cfg_filler_tsc_cmd);
+	install_element(TRX_NODE, &cfg_filler_rach_delay_cmd);
+	install_element(TRX_NODE, &cfg_ctr_error_threshold_cmd);
+	install_element(TRX_NODE, &cfg_no_ctr_error_threshold_cmd);
+	install_element(TRX_NODE, &cfg_stack_size_cmd);
 
 	install_element(TRX_NODE, &cfg_chan_cmd);
 	install_node(&chan_node, dummy_config_write);
