@@ -1,6 +1,8 @@
 /*
 * Copyright 2008, 2009, 2010 Free Software Foundation, Inc.
 *
+* SPDX-License-Identifier: GPL-3.0+
+*
 * This software is distributed under the terms of the GNU Public License.
 * See the COPYING file in the main directory for details.
 *
@@ -22,6 +24,7 @@
 */
 
 #include <stdio.h>
+#include <netinet/in.h>
 #include <iomanip>      // std::setprecision
 #include <fstream>
 #include "Transceiver.h"
@@ -29,6 +32,11 @@
 
 extern "C" {
 #include "osmo_signal.h"
+#include "proto_trxd.h"
+
+#include <osmocom/core/utils.h>
+#include <osmocom/core/socket.h>
+#include <osmocom/core/bits.h>
 }
 
 #ifdef HAVE_CONFIG_H
@@ -115,12 +123,12 @@ Transceiver::Transceiver(int wBasePort,
                          size_t tx_sps, size_t rx_sps, size_t chans,
                          GSM::Time wTransmitLatency,
                          RadioInterface *wRadioInterface,
-                         double wRssiOffset)
+                         double wRssiOffset, int wStackSize)
   : mBasePort(wBasePort), mLocalAddr(TRXAddress), mRemoteAddr(GSMcoreAddress),
-    mClockSocket(TRXAddress, wBasePort, GSMcoreAddress, wBasePort + 100),
-    mTransmitLatency(wTransmitLatency), mRadioInterface(wRadioInterface),
-    rssiOffset(wRssiOffset), sig_cbfn(NULL),
-    mSPSTx(tx_sps), mSPSRx(rx_sps), mChans(chans), mEdge(false), mOn(false), mForceClockInterface(false),
+    mClockSocket(-1), mTransmitLatency(wTransmitLatency), mRadioInterface(wRadioInterface),
+    rssiOffset(wRssiOffset), stackSize(wStackSize),
+    mSPSTx(tx_sps), mSPSRx(rx_sps), mChans(chans), mExtRACH(false), mEdge(false),
+    mOn(false), mForceClockInterface(false),
     mTxFreq(0.0), mRxFreq(0.0), mTSC(0), mMaxExpectedDelayAB(0), mMaxExpectedDelayNB(0),
     mWriteBurstToDiskMask(0)
 {
@@ -139,14 +147,21 @@ Transceiver::~Transceiver()
 
   sigProcLibDestroy();
 
+  if (mClockSocket >= 0)
+    close(mClockSocket);
+
   for (size_t i = 0; i < mChans; i++) {
-    mControlServiceLoopThreads[i]->cancel();
-    mControlServiceLoopThreads[i]->join();
-    delete mControlServiceLoopThreads[i];
+    if (mControlServiceLoopThreads[i]) {
+      mControlServiceLoopThreads[i]->cancel();
+      mControlServiceLoopThreads[i]->join();
+      delete mControlServiceLoopThreads[i];
+    }
 
     mTxPriorityQueues[i].clear();
-    delete mCtrlSockets[i];
-    delete mDataSockets[i];
+    if (mCtrlSockets[i] >= 0)
+      close(mCtrlSockets[i]);
+    if (mDataSockets[i] >= 0)
+      close(mDataSockets[i]);
   }
 }
 
@@ -165,20 +180,20 @@ bool Transceiver::init(FillerType filler, size_t rtsc, unsigned rach_delay,
   int d_srcport, d_dstport, c_srcport, c_dstport;
 
   if (!mChans) {
-    LOG(ALERT) << "No channels assigned";
+    LOG(FATAL) << "No channels assigned";
     return false;
   }
 
   if (!sigProcLibSetup()) {
-    LOG(ALERT) << "Failed to initialize signal processing library";
+    LOG(FATAL) << "Failed to initialize signal processing library";
     return false;
   }
 
   mExtRACH = ext_rach;
   mEdge = edge;
 
-  mDataSockets.resize(mChans);
-  mCtrlSockets.resize(mChans);
+  mDataSockets.resize(mChans, -1);
+  mCtrlSockets.resize(mChans, -1);
   mControlServiceLoopThreads.resize(mChans);
   mTxPriorityQueueServiceLoopThreads.resize(mChans);
   mRxServiceLoopThreads.resize(mChans);
@@ -186,20 +201,39 @@ bool Transceiver::init(FillerType filler, size_t rtsc, unsigned rach_delay,
   mTxPriorityQueues.resize(mChans);
   mReceiveFIFO.resize(mChans);
   mStates.resize(mChans);
+  mVersionTRXD.resize(mChans);
 
   /* Filler table retransmissions - support only on channel 0 */
   if (filler == FILLER_DUMMY)
     mStates[0].mRetrans = true;
 
   /* Setup sockets */
+  mClockSocket = osmo_sock_init2(AF_UNSPEC, SOCK_DGRAM, IPPROTO_UDP,
+				    mLocalAddr.c_str(), mBasePort,
+				    mRemoteAddr.c_str(), mBasePort + 100,
+				    OSMO_SOCK_F_BIND | OSMO_SOCK_F_CONNECT);
+  if (mClockSocket < 0)
+    return false;
+
   for (size_t i = 0; i < mChans; i++) {
     c_srcport = mBasePort + 2 * i + 1;
     c_dstport = mBasePort + 2 * i + 101;
     d_srcport = mBasePort + 2 * i + 2;
     d_dstport = mBasePort + 2 * i + 102;
 
-    mCtrlSockets[i] = new UDPSocket(mLocalAddr.c_str(), c_srcport, mRemoteAddr.c_str(), c_dstport);
-    mDataSockets[i] = new UDPSocket(mLocalAddr.c_str(), d_srcport, mRemoteAddr.c_str(), d_dstport);
+    mCtrlSockets[i] = osmo_sock_init2(AF_UNSPEC, SOCK_DGRAM, IPPROTO_UDP,
+                                      mLocalAddr.c_str(), c_srcport,
+                                      mRemoteAddr.c_str(), c_dstport,
+				      OSMO_SOCK_F_BIND | OSMO_SOCK_F_CONNECT);
+    if (mCtrlSockets[i] < 0)
+      return false;
+
+    mDataSockets[i] = osmo_sock_init2(AF_UNSPEC, SOCK_DGRAM, IPPROTO_UDP,
+                                      mLocalAddr.c_str(), d_srcport,
+                                      mRemoteAddr.c_str(), d_dstport,
+				      OSMO_SOCK_F_BIND | OSMO_SOCK_F_CONNECT);
+    if (mCtrlSockets[i] < 0)
+      return false;
   }
 
   /* Randomize the central clock */
@@ -211,10 +245,12 @@ bool Transceiver::init(FillerType filler, size_t rtsc, unsigned rach_delay,
 
   /* Start control threads */
   for (size_t i = 0; i < mChans; i++) {
-    TransceiverChannel *chan = new TransceiverChannel(this, i);
-    mControlServiceLoopThreads[i] = new Thread(32768);
+    TrxChanThParams *params = (TrxChanThParams *)malloc(sizeof(struct TrxChanThParams));
+    params->trx = this;
+    params->num = i;
+    mControlServiceLoopThreads[i] = new Thread(stackSize);
     mControlServiceLoopThreads[i]->start((void * (*)(void*))
-                                 ControlServiceLoopAdapter, (void*) chan);
+                                 ControlServiceLoopAdapter, (void*) params);
 
     if (i && filler == FILLER_DUMMY)
       filler = FILLER_ZERO;
@@ -225,17 +261,6 @@ bool Transceiver::init(FillerType filler, size_t rtsc, unsigned rach_delay,
   return true;
 }
 
-void Transceiver::setSignalHandler(osmo_signal_cbfn cbfn)
-{
-  if (this->sig_cbfn)
-    osmo_signal_unregister_handler(SS_TRANSC, this->sig_cbfn, NULL);
-
-  if (cbfn) {
-    this->sig_cbfn = cbfn;
-    osmo_signal_register_handler(SS_TRANSC, this->sig_cbfn, NULL);
-  }
-}
-
 /*
  * Start the transceiver
  *
@@ -260,13 +285,13 @@ bool Transceiver::start()
   mLatencyUpdateTime = time;
 
   if (!mRadioInterface->start()) {
-    LOG(ALERT) << "Device failed to start";
+    LOG(FATAL) << "Device failed to start";
     return false;
   }
 
   /* Device is running - launch I/O threads */
-  mRxLowerLoopThread = new Thread(32768);
-  mTxLowerLoopThread = new Thread(32768);
+  mRxLowerLoopThread = new Thread(stackSize);
+  mTxLowerLoopThread = new Thread(stackSize);
   mTxLowerLoopThread->start((void * (*)(void*))
                             TxLowerLoopAdapter,(void*) this);
   mRxLowerLoopThread->start((void * (*)(void*))
@@ -274,15 +299,19 @@ bool Transceiver::start()
 
   /* Launch uplink and downlink burst processing threads */
   for (size_t i = 0; i < mChans; i++) {
-    TransceiverChannel *chan = new TransceiverChannel(this, i);
-    mRxServiceLoopThreads[i] = new Thread(32768);
+    TrxChanThParams *params = (TrxChanThParams *)malloc(sizeof(struct TrxChanThParams));
+    params->trx = this;
+    params->num = i;
+    mRxServiceLoopThreads[i] = new Thread(stackSize);
     mRxServiceLoopThreads[i]->start((void * (*)(void*))
-                            RxUpperLoopAdapter, (void*) chan);
+                            RxUpperLoopAdapter, (void*) params);
 
-    chan = new TransceiverChannel(this, i);
-    mTxPriorityQueueServiceLoopThreads[i] = new Thread(32768);
+    params = (TrxChanThParams *)malloc(sizeof(struct TrxChanThParams));
+    params->trx = this;
+    params->num = i;
+    mTxPriorityQueueServiceLoopThreads[i] = new Thread(stackSize);
     mTxPriorityQueueServiceLoopThreads[i]->start((void * (*)(void*))
-                            TxUpperLoopAdapter, (void*) chan);
+                            TxUpperLoopAdapter, (void*) params);
   }
 
   mForceClockInterface = true;
@@ -342,12 +371,12 @@ void Transceiver::addRadioVector(size_t chan, BitVector &bits,
   radioVector *radio_burst;
 
   if (chan >= mTxPriorityQueues.size()) {
-    LOG(ALERT) << "Invalid channel " << chan;
+    LOGCHAN(chan, DTRXDDL, FATAL) << "Invalid channel";
     return;
   }
 
   if (wTime.TN() > 7) {
-    LOG(ALERT) << "Received burst with invalid slot " << wTime.TN();
+    LOGCHAN(chan, DTRXDDL, FATAL) << "Received burst with invalid slot " << wTime.TN();
     return;
   }
 
@@ -390,7 +419,7 @@ void Transceiver::pushRadioVector(GSM::Time &nowTime)
     state = &mStates[i];
 
     while ((burst = mTxPriorityQueues[i].getStaleBurst(nowTime))) {
-      LOGCHAN(i, DMAIN, NOTICE) << "dumping STALE burst in TRX->SDR interface ("
+      LOGCHAN(i, DTRXDDL, NOTICE) << "dumping STALE burst in TRX->SDR interface ("
                   << burst->getTime() <<" vs " << nowTime << "), retrans=" << state->mRetrans;
       if (state->mRetrans)
         updateFillerTable(i, burst);
@@ -529,8 +558,8 @@ CorrType Transceiver::expectedCorrType(GSM::Time currTime,
       return mExtRACH ? EXT_RACH : RACH;
     else if ((mod52 == 25) || (mod52 == 51))
       return IDLE;
-    else
-      return TSC;
+    else /* Enable 8-PSK burst detection if EDGE is enabled */
+      return mEdge ? EDGE : TSC;
     break;
   }
   case LOOPBACK:
@@ -558,44 +587,57 @@ void writeToFile(radioVector *radio_burst, size_t chan)
 /*
  * Pull bursts from the FIFO and handle according to the slot
  * and burst correlation type. Equalzation is currently disabled.
+ * returns 0 on success (bi filled), negative on error (bi content undefined):
+ *        -ENOENT: timeslot is off (fn and tn in bi are filled),
+ *        -EIO: read error
  */
-SoftVector *Transceiver::pullRadioVector(GSM::Time &wTime, double &RSSI, bool &isRssiValid,
-                                         double &timingOffset, double &noise,
-                                         size_t chan)
+int Transceiver::pullRadioVector(size_t chan, struct trx_ul_burst_ind *bi)
 {
   int rc;
-  complex amp;
-  float toa, max = -1.0, avg = 0.0;
+  struct estim_burst_params ebp;
+  float max = -1.0, avg = 0.0;
+  unsigned max_toa;
   int max_i = -1;
   signalVector *burst;
-  SoftVector *bits = NULL;
+  GSM::Time burstTime;
+  SoftVector *rxBurst;
   TransceiverState *state = &mStates[chan];
-  isRssiValid = false;
 
   /* Blocking FIFO read */
   radioVector *radio_burst = mReceiveFIFO[chan]->read();
-  if (!radio_burst)
-    return NULL;
+  if (!radio_burst) {
+    LOGCHAN(chan, DTRXDUL, ERROR) << "ReceiveFIFO->read() returned no burst";
+    return -EIO;
+  }
 
   /* Set time and determine correlation type */
-  GSM::Time time = radio_burst->getTime();
-  CorrType type = expectedCorrType(time, chan);
+  burstTime = radio_burst->getTime();
+  CorrType type = expectedCorrType(burstTime, chan);
 
-  /* Enable 8-PSK burst detection if EDGE is enabled */
-  if (mEdge && (type == TSC))
-    type = EDGE;
+  /* Initialize struct bi */
+  bi->nbits = 0;
+  bi->fn = burstTime.FN();
+  bi->tn = burstTime.TN();
+  bi->rssi = 0.0;
+  bi->toa = 0.0;
+  bi->noise = 0.0;
+  bi->idle = false;
+  bi->modulation = MODULATION_GMSK;
+  bi->tss = 0; /* TODO: we only support tss 0 right now */
+  bi->tsc = 0;
+  bi->ci = 0.0;
 
   /* Debug: dump bursts to disk */
   /* bits 0-7  - chan 0 timeslots
    * bits 8-15 - chan 1 timeslots */
-  if (mWriteBurstToDiskMask & ((1<<time.TN()) << (8*chan)))
+  if (mWriteBurstToDiskMask & ((1<<bi->tn) << (8*chan)))
     writeToFile(radio_burst, chan);
 
   /* No processing if the timeslot is off.
    * Not even power level or noise calculation. */
   if (type == OFF) {
     delete radio_burst;
-    return NULL;
+    return -ENOENT;
   }
 
   /* Select the diversity channel with highest energy */
@@ -609,59 +651,65 @@ SoftVector *Transceiver::pullRadioVector(GSM::Time &wTime, double &RSSI, bool &i
   }
 
   if (max_i < 0) {
-    LOG(ALERT) << "Received empty burst";
-    delete radio_burst;
-    return NULL;
+    LOGCHAN(chan, DTRXDUL, FATAL) << "Received empty burst";
+    goto ret_idle;
   }
 
   /* Average noise on diversity paths and update global levels */
   burst = radio_burst->getVector(max_i);
   avg = sqrt(avg / radio_burst->chans());
 
-  wTime = time;
-  RSSI = 20.0 * log10(rxFullScale / avg);
-
-  /* RSSI estimation are valid */
-  isRssiValid = true;
-
   if (type == IDLE) {
     /* Update noise levels */
     state->mNoises.insert(avg);
     state->mNoiseLev = state->mNoises.avg();
-    noise = 20.0 * log10(rxFullScale / state->mNoiseLev);
-
-    delete radio_burst;
-    return NULL;
-  } else {
-    /* Do not update noise levels */
-    noise = 20.0 * log10(rxFullScale / state->mNoiseLev);
   }
 
-  unsigned max_toa = (type == RACH || type == EXT_RACH) ?
-                      mMaxExpectedDelayAB : mMaxExpectedDelayNB;
+  bi->rssi = 20.0 * log10(rxFullScale / avg) + rssiOffset;
+  bi->noise = 20.0 * log10(rxFullScale / state->mNoiseLev) + rssiOffset;
+
+  if (type == IDLE)
+    goto ret_idle;
+
+  max_toa = (type == RACH || type == EXT_RACH) ?
+            mMaxExpectedDelayAB : mMaxExpectedDelayNB;
 
   /* Detect normal or RACH bursts */
-  rc = detectAnyBurst(*burst, mTSC, BURST_THRESH, mSPSRx, type, amp, toa, max_toa);
-
-  if (rc > 0) {
-    type = (CorrType) rc;
-  } else if (rc <= 0) {
-    if (rc == -SIGERR_CLIP) {
-      LOG(WARNING) << "Clipping detected on received RACH or Normal Burst";
-    } else if (rc != SIGERR_NONE) {
-      LOG(WARNING) << "Unhandled RACH or Normal Burst detection error";
-    }
-
-    delete radio_burst;
-    return NULL;
+  rc = detectAnyBurst(*burst, mTSC, BURST_THRESH, mSPSRx, type, max_toa, &ebp);
+  if (rc <= 0) {
+    if (rc == -SIGERR_CLIP)
+      LOGCHAN(chan, DTRXDUL, NOTICE) << "Clipping detected on received RACH or Normal Burst";
+    else if (rc != SIGERR_NONE)
+      LOGCHAN(chan, DTRXDUL, NOTICE) << "Unhandled RACH or Normal Burst detection error";
+    goto ret_idle;
   }
 
-  timingOffset = toa;
+  type = (CorrType) rc;
+  bi->toa = ebp.toa;
+  bi->tsc = ebp.tsc;
+  bi->ci = ebp.ci;
+  rxBurst = demodAnyBurst(*burst, mSPSRx, ebp.amp, ebp.toa, type);
 
-  bits = demodAnyBurst(*burst, mSPSRx, amp, toa, type);
+  /* EDGE demodulator returns 444 (gSlotLen * 3) bits */
+  if (rxBurst->size() == EDGE_BURST_NBITS) {
+    bi->modulation = MODULATION_8PSK;
+    bi->nbits = EDGE_BURST_NBITS;
+  } else { /* size() here is actually gSlotLen + 8, due to guard periods */
+    bi->modulation = MODULATION_GMSK;
+    bi->nbits = gSlotLen;
+  }
 
+  // Convert -1..+1 soft bits to 0..1 soft bits
+  vectorSlicer(bi->rx_burst, rxBurst->begin(), bi->nbits);
+
+  delete rxBurst;
   delete radio_burst;
-  return bits;
+  return 0;
+
+ret_idle:
+  bi->idle = true;
+  delete radio_burst;
+  return 0;
 }
 
 void Transceiver::reset()
@@ -702,7 +750,7 @@ static bool match_cmd(char *buf,
   return true;
 }
 
-void Transceiver::driveControl(size_t chan)
+bool Transceiver::driveControl(size_t chan)
 {
   char buffer[MAX_PACKET_LENGTH + 1];
   char response[MAX_PACKET_LENGTH + 1];
@@ -710,17 +758,19 @@ void Transceiver::driveControl(size_t chan)
   int msgLen;
 
   /* Attempt to read from control socket */
-  msgLen = mCtrlSockets[chan]->read(buffer, MAX_PACKET_LENGTH);
-  if (msgLen < 1)
-    return;
+  msgLen = read(mCtrlSockets[chan], buffer, MAX_PACKET_LENGTH);
+  if (msgLen <= 0) {
+    LOGCHAN(chan, DTRXCTRL, NOTICE) << "mCtrlSockets read(" << mCtrlSockets[chan] << ") failed: " << msgLen;
+    return false;
+  }
 
   /* Zero-terminate received string */
   buffer[msgLen] = '\0';
 
   /* Verify a command signature */
   if (strncmp(buffer, "CMD ", 4)) {
-    LOGC(DTRXCTRL, WARNING) << "bogus message on control interface";
-    return;
+    LOGCHAN(chan, DTRXCTRL, NOTICE) << "bogus message on control interface";
+    return false;
   }
 
   /* Set command pointer */
@@ -744,7 +794,7 @@ void Transceiver::driveControl(size_t chan)
     unsigned ts = 0, ss = 0;
     sscanf(params, "%u %u", &ts, &ss);
     if (ts > 7 || ss > 7) {
-      sprintf(response, "RSP NOHANDOVER 1 %u %u", ts, ss);
+      sprintf(response, "RSP HANDOVER 1 %u %u", ts, ss);
     } else {
       mHandover[ts][ss] = true;
       sprintf(response, "RSP HANDOVER 0 %u %u", ts, ss);
@@ -804,7 +854,7 @@ void Transceiver::driveControl(size_t chan)
     sscanf(params, "%d", &freqKhz);
     mRxFreq = freqKhz * 1e3;
     if (!mRadioInterface->tuneRx(mRxFreq, chan)) {
-       LOGC(DTRXCTRL, ALERT) << "RX failed to tune";
+       LOGCHAN(chan, DTRXCTRL, FATAL) << "RX failed to tune";
        sprintf(response,"RSP RXTUNE 1 %d",freqKhz);
     }
     else
@@ -815,7 +865,7 @@ void Transceiver::driveControl(size_t chan)
     sscanf(params, "%d", &freqKhz);
     mTxFreq = freqKhz * 1e3;
     if (!mRadioInterface->tuneTx(mTxFreq, chan)) {
-       LOGC(DTRXCTRL, ALERT) << "TX failed to tune";
+       LOGCHAN(chan, DTRXCTRL, FATAL) << "TX failed to tune";
        sprintf(response,"RSP TXTUNE 1 %d",freqKhz);
     }
     else
@@ -837,146 +887,173 @@ void Transceiver::driveControl(size_t chan)
     int  timeslot;
     sscanf(params, "%d %d", &timeslot, &corrCode);
     if ((timeslot < 0) || (timeslot > 7)) {
-      LOGC(DTRXCTRL, WARNING) << "bogus message on control interface";
+      LOGCHAN(chan, DTRXCTRL, NOTICE) << "bogus message on control interface";
       sprintf(response,"RSP SETSLOT 1 %d %d",timeslot,corrCode);
-      return;
+      return true;
     }
     mStates[chan].chanType[timeslot] = (ChannelCombination) corrCode;
     setModulus(timeslot, chan);
     sprintf(response,"RSP SETSLOT 0 %d %d",timeslot,corrCode);
+  } else if (match_cmd(command, "SETFORMAT", &params)) {
+    // set TRXD protocol version
+    unsigned version_recv;
+    sscanf(params, "%u", &version_recv);
+    LOGCHAN(chan, DTRXCTRL, INFO) << "BTS requests TRXD version switch: " << version_recv;
+    if (version_recv > TRX_DATA_FORMAT_VER) {
+      LOGCHAN(chan, DTRXCTRL, INFO) << "rejecting TRXD version " << version_recv
+                                    << "in favor of " <<  TRX_DATA_FORMAT_VER;
+      sprintf(response, "RSP SETFORMAT %u %u", TRX_DATA_FORMAT_VER, version_recv);
+    } else {
+      LOGCHAN(chan, DTRXCTRL, NOTICE) << "switching to TRXD version " << version_recv;
+      mVersionTRXD[chan] = version_recv;
+      sprintf(response, "RSP SETFORMAT %u %u", version_recv, version_recv);
+    }
   } else if (match_cmd(command, "_SETBURSTTODISKMASK", &params)) {
-    // debug command! may change or disapear without notice
+    // debug command! may change or disappear without notice
     // set a mask which bursts to dump to disk
     int mask;
     sscanf(params, "%d", &mask);
     mWriteBurstToDiskMask = mask;
     sprintf(response,"RSP _SETBURSTTODISKMASK 0 %d",mask);
   } else {
-    LOGC(DTRXCTRL, WARNING) << "bogus command " << command << " on control interface.";
+    LOGCHAN(chan, DTRXCTRL, NOTICE) << "bogus command " << command << " on control interface.";
     sprintf(response,"RSP ERR 1");
   }
 
   LOGCHAN(chan, DTRXCTRL, INFO) << "response is '" << response << "'";
-  mCtrlSockets[chan]->write(response, strlen(response) + 1);
+  msgLen = write(mCtrlSockets[chan], response, strlen(response) + 1);
+  if (msgLen <= 0) {
+    LOGCHAN(chan, DTRXCTRL, NOTICE) << "mCtrlSockets write(" << mCtrlSockets[chan] << ") failed: " << msgLen;
+    return false;
+  }
+  return true;
 }
 
 bool Transceiver::driveTxPriorityQueue(size_t chan)
 {
+  int msgLen;
   int burstLen;
-  char buffer[EDGE_BURST_NBITS + 50];
+  struct trxd_hdr_v01_dl *dl;
+  char buffer[sizeof(*dl) + EDGE_BURST_NBITS];
+  uint32_t fn;
 
   // check data socket
-  size_t msgLen = mDataSockets[chan]->read(buffer, sizeof(buffer));
-
-  if (msgLen == gSlotLen + 1 + 4 + 1) {
-    burstLen = gSlotLen;
-  } else if (msgLen == EDGE_BURST_NBITS + 1 + 4 + 1) {
-    if (mSPSTx != 4)
-      return false;
-
-    burstLen = EDGE_BURST_NBITS;
-  } else {
-    LOG(ERR) << "badly formatted packet on GSM->TRX interface";
+  msgLen = read(mDataSockets[chan], buffer, sizeof(buffer));
+  if (msgLen <= 0) {
+    LOGCHAN(chan, DTRXDDL, NOTICE) << "mDataSockets read(" << mCtrlSockets[chan] << ") failed: " << msgLen;
     return false;
   }
 
-  int timeSlot = (int) buffer[0];
-  uint64_t frameNum = 0;
-  for (int i = 0; i < 4; i++)
-    frameNum = (frameNum << 8) | (0x0ff & buffer[i+1]);
+  switch (msgLen) {
+    case sizeof(*dl) + gSlotLen: /* GSM burst */
+      burstLen = gSlotLen;
+      break;
+    case sizeof(*dl) + EDGE_BURST_NBITS: /* EDGE burst */
+      if (mSPSTx != 4) {
+        LOGCHAN(chan, DTRXDDL, ERROR) << "EDGE burst received but SPS is set to " << mSPSTx;
+        return false;
+      }
+      burstLen = EDGE_BURST_NBITS;
+      break;
+    default:
+      LOGCHAN(chan, DTRXDDL, ERROR) << "badly formatted packet on GSM->TRX interface (len="<< msgLen << ")";
+      return false;
+  }
 
-  LOG(DEBUG) << "rcvd. burst at: " << GSM::Time(frameNum,timeSlot);
+  dl = (struct trxd_hdr_v01_dl *) buffer;
+
+  /* Convert TDMA FN to the host endianness */
+  fn = osmo_load32be(&dl->common.fn);
+
+  /* Make sure we support the received header format */
+  switch (dl->common.version) {
+  case 0:
+  /* Version 1 has the same format */
+  case 1:
+    break;
+  default:
+    LOGCHAN(chan, DTRXDDL, ERROR) << "Rx TRXD message with unknown header version " << unsigned(dl->common.version);
+    return false;
+  }
+
+  LOGCHAN(chan, DTRXDDL, DEBUG) << "Rx TRXD message (hdr_ver=" << unsigned(dl->common.version)
+    << "): fn=" << fn << ", tn=" << unsigned(dl->common.tn) << ", burst_len=" << burstLen;
 
-  int RSSI = (int) buffer[5];
   BitVector newBurst(burstLen);
   BitVector::iterator itr = newBurst.begin();
-  char *bufferItr = buffer+6;
+  uint8_t *bufferItr = dl->soft_bits;
   while (itr < newBurst.end())
     *itr++ = *bufferItr++;
 
-  GSM::Time currTime = GSM::Time(frameNum,timeSlot);
+  GSM::Time currTime = GSM::Time(fn, dl->common.tn);
 
-  addRadioVector(chan, newBurst, RSSI, currTime);
+  addRadioVector(chan, newBurst, dl->tx_att, currTime);
 
   return true;
-
-
 }
 
-void Transceiver::driveReceiveRadio()
+bool Transceiver::driveReceiveRadio()
 {
   int rc = mRadioInterface->driveReceiveRadio();
   if (rc == 0) {
     usleep(100000);
-  } else if (rc < 0) {
-    LOG(FATAL) << "radio Interface receive failed, requesting stop.";
-    osmo_signal_dispatch(SS_TRANSC, S_TRANSC_STOP_REQUIRED, this);
-  } else if (mForceClockInterface || mTransmitDeadlineClock > mLastClockUpdateTime + GSM::Time(216,0)) {
-    mForceClockInterface = false;
-    writeClockInterface();
+    return true;
   }
+  if (rc < 0)
+    return false;
+
+  if (mForceClockInterface || mTransmitDeadlineClock > mLastClockUpdateTime + GSM::Time(216,0)) {
+    mForceClockInterface = false;
+    return writeClockInterface();
+  }
+  return true;
 }
 
-void Transceiver::logRxBurst(size_t chan, SoftVector *burst, GSM::Time time, double dbm,
-                             double rssi, double noise, double toa)
+void Transceiver::logRxBurst(size_t chan, const struct trx_ul_burst_ind *bi)
 {
-  LOG(DEBUG) << std::fixed << std::right
-    << " chan: "   << chan
-    << " time: "   << time
-    << " RSSI: "   << std::setw(5) << std::setprecision(1) << rssi
-                   << "dBFS/" << std::setw(6) << -dbm << "dBm"
-    << " noise: "  << std::setw(5) << std::setprecision(1) << noise
-                   << "dBFS/" << std::setw(6) << -(noise + rssiOffset) << "dBm"
-    << " TOA: "    << std::setw(5) << std::setprecision(2) << toa
-    << " bits: "   << *burst;
+  std::ostringstream os;
+  for (size_t i=0; i < bi->nbits; i++) {
+    if (bi->rx_burst[i] > 0.5) os << "1";
+    else if (bi->rx_burst[i] > 0.25) os << "|";
+    else if (bi->rx_burst[i] > 0.0) os << "'";
+    else os << "-";
+  }
+
+  LOGCHAN(chan, DTRXDUL, DEBUG) << std::fixed << std::right
+    << " time: "   << unsigned(bi->tn) << ":" << bi->fn
+    << " RSSI: "   << std::setw(5) << std::setprecision(1) << (bi->rssi - rssiOffset)
+                   << "dBFS/" << std::setw(6) << -bi->rssi << "dBm"
+    << " noise: "  << std::setw(5) << std::setprecision(1) << (bi->noise - rssiOffset)
+                   << "dBFS/" << std::setw(6) << -bi->noise << "dBm"
+    << " TOA: "    << std::setw(5) << std::setprecision(2) << bi->toa
+    << " C/I: "    << std::setw(5) << std::setprecision(2) << bi->ci << "dB"
+    << " bits: "   << os;
 }
 
-void Transceiver::driveReceiveFIFO(size_t chan)
+bool Transceiver::driveReceiveFIFO(size_t chan)
 {
-  SoftVector *rxBurst = NULL;
-  double RSSI; // in dBFS
-  double dBm;  // in dBm
-  double TOA;  // in symbols
-  int TOAint;  // in 1/256 symbols
-  double noise; // noise level in dBFS
-  GSM::Time burstTime;
-  bool isRssiValid; // are RSSI, noise and burstTime valid
-  unsigned nbits = gSlotLen;
+  struct trx_ul_burst_ind bi;
+  int rc;
 
-  rxBurst = pullRadioVector(burstTime, RSSI, isRssiValid, TOA, noise, chan);
-  if (!rxBurst)
-    return;
+  if ((rc = pullRadioVector(chan, &bi)) < 0) {
+    if (rc == -ENOENT) { /* timeslot off, continue processing */
+      LOGCHAN(chan, DTRXDUL, DEBUG) << unsigned(bi.tn) << ":" << bi.fn << " timeslot is off";
+      return true;
+    }
+    return false; /* other errors: we want to stop the process */
+  }
 
-  // Convert -1..+1 soft bits to 0..1 soft bits
-  vectorSlicer(rxBurst);
+  if (!bi.idle)
+    logRxBurst(chan, &bi);
 
-  /*
-   * EDGE demodulator returns 444 (148 * 3) bits
-   */
-  if (rxBurst->size() == gSlotLen * 3)
-    nbits = gSlotLen * 3;
-
-  dBm = RSSI + rssiOffset;
-  logRxBurst(chan, rxBurst, burstTime, dBm, RSSI, noise, TOA);
-
-  TOAint = (int) (TOA * 256.0 + 0.5); // round to closest integer
-
-  char burstString[nbits + 10];
-  burstString[0] = burstTime.TN();
-  for (int i = 0; i < 4; i++)
-    burstString[1+i] = (burstTime.FN() >> ((3-i)*8)) & 0x0ff;
-  burstString[5] = (int)dBm;
-  burstString[6] = (TOAint >> 8) & 0x0ff;
-  burstString[7] = TOAint & 0x0ff;
-  SoftVector::iterator burstItr = rxBurst->begin();
-
-  for (unsigned i = 0; i < nbits; i++)
-    burstString[8 + i] = (char) round((*burstItr++) * 255.0);
-
-  burstString[nbits + 9] = '\0';
-  delete rxBurst;
-
-  mDataSockets[chan]->write(burstString, nbits + 10);
+  switch (mVersionTRXD[chan]) {
+    case 0:
+      return trxd_send_burst_ind_v0(chan, mDataSockets[chan], &bi);
+    case 1:
+      return trxd_send_burst_ind_v1(chan, mDataSockets[chan], &bi);
+    default:
+      OSMO_ASSERT(false);
+  }
 }
 
 void Transceiver::driveTxFIFO()
@@ -997,7 +1074,7 @@ void Transceiver::driveTxFIFO()
 
   if (mOn) {
     //radioClock->wait(); // wait until clock updates
-    LOG(DEBUG) << "radio clock " << radioClock->get();
+    LOGC(DTRXCLK, DEBUG) << "radio clock " << radioClock->get();
     while (radioClock->get() + mTransmitLatency > mTransmitDeadlineClock) {
       // if underrun, then we're not providing bursts to radio/USRP fast
       //   enough.  Need to increase latency by one GSM frame.
@@ -1006,8 +1083,9 @@ void Transceiver::driveTxFIFO()
           // only update latency at the defined frame interval
           if (radioClock->get() > mLatencyUpdateTime + GSM::Time(USB_LATENCY_INTRVL)) {
             mTransmitLatency = mTransmitLatency + GSM::Time(1,0);
-            LOG(INFO) << "new latency: " << mTransmitLatency << " (underrun "
-                      << radioClock->get() << " vs " << mLatencyUpdateTime + GSM::Time(USB_LATENCY_INTRVL) << ")";
+            LOGC(DTRXCLK, INFO) << "new latency: " << mTransmitLatency << " (underrun "
+                                << radioClock->get() << " vs "
+                                << mLatencyUpdateTime + GSM::Time(USB_LATENCY_INTRVL) << ")";
             mLatencyUpdateTime = radioClock->get();
           }
         }
@@ -1017,7 +1095,7 @@ void Transceiver::driveTxFIFO()
           if (mTransmitLatency > mRadioInterface->minLatency()) {
               if (radioClock->get() > mLatencyUpdateTime + GSM::Time(216,0)) {
               mTransmitLatency.decTN();
-              LOG(INFO) << "reduced latency: " << mTransmitLatency;
+              LOGC(DTRXCLK, INFO) << "reduced latency: " << mTransmitLatency;
               mLatencyUpdateTime = radioClock->get();
             }
           }
@@ -1034,35 +1112,42 @@ void Transceiver::driveTxFIFO()
 
 
 
-void Transceiver::writeClockInterface()
+bool Transceiver::writeClockInterface()
 {
+  int msgLen;
   char command[50];
   // FIXME -- This should be adaptive.
   sprintf(command,"IND CLOCK %llu",(unsigned long long) (mTransmitDeadlineClock.FN()+2));
 
-  LOG(INFO) << "ClockInterface: sending " << command;
+  LOGC(DTRXCLK, INFO) << "sending " << command;
 
-  mClockSocket.write(command, strlen(command) + 1);
+  msgLen = write(mClockSocket, command, strlen(command) + 1);
+  if (msgLen <= 0) {
+    LOGC(DTRXCLK, ERROR) << "mClockSocket write(" << mClockSocket << ") failed: " << msgLen;
+    return false;
+  }
 
   mLastClockUpdateTime = mTransmitDeadlineClock;
-
+  return true;
 }
 
-void *RxUpperLoopAdapter(TransceiverChannel *chan)
+void *RxUpperLoopAdapter(TrxChanThParams *params)
 {
   char thread_name[16];
-  Transceiver *trx = chan->trx;
-  size_t num = chan->num;
+  Transceiver *trx = params->trx;
+  size_t num = params->num;
 
-  delete chan;
+  free(params);
 
   snprintf(thread_name, 16, "RxUpper%zu", num);
   set_selfthread_name(thread_name);
 
-  trx->setPriority(0.42);
-
   while (1) {
-    trx->driveReceiveFIFO(num);
+    if (!trx->driveReceiveFIFO(num)) {
+      LOGCHAN(num, DTRXDUL, FATAL) << "Something went wrong in thread " << thread_name << ", requesting stop";
+      osmo_signal_dispatch(SS_MAIN, S_MAIN_STOP_REQUIRED, NULL);
+      break;
+    }
     pthread_testcancel();
   }
   return NULL;
@@ -1072,10 +1157,12 @@ void *RxLowerLoopAdapter(Transceiver *transceiver)
 {
   set_selfthread_name("RxLower");
 
-  transceiver->setPriority(0.45);
-
   while (1) {
-    transceiver->driveReceiveRadio();
+    if (!transceiver->driveReceiveRadio()) {
+      LOGC(DTRXDUL, FATAL) << "Something went wrong in thread RxLower, requesting stop";
+      osmo_signal_dispatch(SS_MAIN, S_MAIN_STOP_REQUIRED, NULL);
+      break;
+    }
     pthread_testcancel();
   }
   return NULL;
@@ -1085,8 +1172,6 @@ void *TxLowerLoopAdapter(Transceiver *transceiver)
 {
   set_selfthread_name("TxLower");
 
-  transceiver->setPriority(0.44);
-
   while (1) {
     transceiver->driveTxFIFO();
     pthread_testcancel();
@@ -1094,39 +1179,45 @@ void *TxLowerLoopAdapter(Transceiver *transceiver)
   return NULL;
 }
 
-void *ControlServiceLoopAdapter(TransceiverChannel *chan)
+void *ControlServiceLoopAdapter(TrxChanThParams *params)
 {
   char thread_name[16];
-  Transceiver *trx = chan->trx;
-  size_t num = chan->num;
+  Transceiver *trx = params->trx;
+  size_t num = params->num;
 
-  delete chan;
+  free(params);
 
   snprintf(thread_name, 16, "CtrlService%zu", num);
   set_selfthread_name(thread_name);
 
   while (1) {
-    trx->driveControl(num);
+    if (!trx->driveControl(num)) {
+      LOGCHAN(num, DTRXCTRL, FATAL) << "Something went wrong in thread " << thread_name << ", requesting stop";
+      osmo_signal_dispatch(SS_MAIN, S_MAIN_STOP_REQUIRED, NULL);
+      break;
+    }
     pthread_testcancel();
   }
   return NULL;
 }
 
-void *TxUpperLoopAdapter(TransceiverChannel *chan)
+void *TxUpperLoopAdapter(TrxChanThParams *params)
 {
   char thread_name[16];
-  Transceiver *trx = chan->trx;
-  size_t num = chan->num;
+  Transceiver *trx = params->trx;
+  size_t num = params->num;
 
-  delete chan;
+  free(params);
 
   snprintf(thread_name, 16, "TxUpper%zu", num);
   set_selfthread_name(thread_name);
 
-  trx->setPriority(0.40);
-
   while (1) {
-    trx->driveTxPriorityQueue(num);
+    if (!trx->driveTxPriorityQueue(num)) {
+      LOGCHAN(num, DTRXDDL, FATAL) << "Something went wrong in thread " << thread_name << ", requesting stop";
+      osmo_signal_dispatch(SS_MAIN, S_MAIN_STOP_REQUIRED, NULL);
+      break;
+    }
     pthread_testcancel();
   }
   return NULL;
