@@ -1,6 +1,8 @@
 /*
 * Copyright 2008 Free Software Foundation, Inc.
 *
+* SPDX-License-Identifier: GPL-3.0+
+*
 * This software is distributed under the terms of the GNU Public License.
 * See the COPYING file in the main directory for details.
 *
@@ -25,7 +27,6 @@
 #include "radioInterface.h"
 #include "Interthread.h"
 #include "GSMCommon.h"
-#include "Sockets.h"
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -38,19 +39,9 @@ extern "C" {
 class Transceiver;
 
 /** Channel descriptor for transceiver object and channel number pair */
-struct TransceiverChannel {
-  TransceiverChannel(Transceiver *trx, int num)
-  {
-    this->trx = trx;
-    this->num = num;
-  }
-
-  ~TransceiverChannel()
-  {
-  }
-
-  Transceiver *trx;
-  size_t num;
+struct TrxChanThParams {
+	Transceiver *trx;
+	size_t num;
 };
 
 /** Internal transceiver state variables */
@@ -108,7 +99,7 @@ public:
               size_t tx_sps, size_t rx_sps, size_t chans,
               GSM::Time wTransmitLatency,
               RadioInterface *wRadioInterface,
-              double wRssiOffset);
+              double wRssiOffset, int stackSize);
 
   /** Destructor */
   ~Transceiver();
@@ -130,8 +121,6 @@ public:
   /** accessor for number of channels */
   size_t numChans() const { return mChans; };
 
-  void setSignalHandler(osmo_signal_cbfn cbfn);
-
   /** Codes for channel combinations */
   typedef enum {
     FILL,               ///< Channel is transmitted, but unused
@@ -157,9 +146,9 @@ private:
   std::string mLocalAddr;
   std::string mRemoteAddr;
 
-  std::vector<UDPSocket *> mDataSockets;  ///< socket for writing to/reading from GSM core
-  std::vector<UDPSocket *> mCtrlSockets;  ///< socket for writing/reading control commands from GSM core
-  UDPSocket mClockSocket;                 ///< socket for writing clock updates to GSM core
+  std::vector<int> mDataSockets;  ///< socket for writing to/reading from GSM core
+  std::vector<int> mCtrlSockets;  ///< socket for writing/reading control commands from GSM core
+  int mClockSocket;               ///< socket for writing clock updates to GSM core
 
   std::vector<VectorQueue> mTxPriorityQueues;   ///< priority queue of transmit bursts received from GSM core
   std::vector<VectorFIFO *>  mReceiveFIFO;      ///< radioInterface FIFO of receive bursts
@@ -180,8 +169,7 @@ private:
   double rxFullScale;                     ///< full scale output to radio
 
   double rssiOffset;                      ///< RSSI to dBm conversion offset
-
-  osmo_signal_cbfn *sig_cbfn;              ///< Registered Signal Handler to announce events.
+  int stackSize;                      ///< stack size for threads, 0 = OS default
 
   /** modulate and add a burst to the transmit queue */
   void addRadioVector(size_t chan, BitVector &bits,
@@ -194,9 +182,7 @@ private:
   void pushRadioVector(GSM::Time &nowTime);
 
   /** Pull and demodulate a burst from the receive FIFO */
-  SoftVector *pullRadioVector(GSM::Time &wTime, double &RSSI, bool &isRssiValid,
-                              double &timingOffset, double &noise,
-                              size_t chan = 0);
+  int pullRadioVector(size_t chan, struct trx_ul_burst_ind *ind);
 
   /** Set modulus for specific timeslot */
   void setModulus(size_t timeslot, size_t chan);
@@ -205,7 +191,7 @@ private:
   CorrType expectedCorrType(GSM::Time currTime, size_t chan);
 
   /** send messages over the clock socket */
-  void writeClockInterface(void);
+  bool writeClockInterface(void);
 
   int mSPSTx;                          ///< number of samples per Tx symbol
   int mSPSRx;                          ///< number of samples per Rx symbol
@@ -223,27 +209,28 @@ private:
   unsigned mMaxExpectedDelayNB;        ///< maximum expected time-of-arrival offset in GSM symbols for Normal Bursts
   unsigned mWriteBurstToDiskMask;      ///< debug: bitmask to indicate which timeslots to dump to disk
 
+  std::vector<unsigned> mVersionTRXD;  ///< Format version to use for TRXD protocol communication, per channel
   std::vector<TransceiverState> mStates;
 
   /** Start and stop I/O threads through the control socket API */
   bool start();
   void stop();
 
-  /** Protect destructor accessable stop call */
+  /** Protect destructor accessible stop call */
   Mutex mLock;
 
 protected:
   /** drive lower receive I/O and burst generation */
-  void driveReceiveRadio();
+  bool driveReceiveRadio();
 
   /** drive demodulation of GSM bursts */
-  void driveReceiveFIFO(size_t chan);
+  bool driveReceiveFIFO(size_t chan);
 
   /** drive transmission of GSM bursts */
   void driveTxFIFO();
 
   /** drive handling of control messages from GSM core */
-  void driveControl(size_t chan);
+  bool driveControl(size_t chan);
 
   /**
     drive modulation and sorting of GSM bursts from GSM core
@@ -251,34 +238,26 @@ protected:
   */
   bool driveTxPriorityQueue(size_t chan);
 
-  friend void *RxUpperLoopAdapter(TransceiverChannel *);
-
-  friend void *TxUpperLoopAdapter(TransceiverChannel *);
-
-  friend void *RxLowerLoopAdapter(Transceiver *);
-
-  friend void *TxLowerLoopAdapter(Transceiver *);
-
-  friend void *ControlServiceLoopAdapter(TransceiverChannel *);
+  friend void *RxUpperLoopAdapter(TrxChanThParams *params);
+  friend void *TxUpperLoopAdapter(TrxChanThParams *params);
+  friend void *RxLowerLoopAdapter(Transceiver *transceiver);
+  friend void *TxLowerLoopAdapter(Transceiver *transceiver);
+  friend void *ControlServiceLoopAdapter(TrxChanThParams *params);
 
 
   void reset();
 
-  /** set priority on current thread */
-  void setPriority(float prio = 0.5) { mRadioInterface->setPriority(prio); }
-
-  void logRxBurst(size_t chan, SoftVector *burst, GSM::Time time, double dbm,
-                  double rssi, double noise, double toa);
+  void logRxBurst(size_t chan, const struct trx_ul_burst_ind *bi);
 };
 
-void *RxUpperLoopAdapter(TransceiverChannel *);
+void *RxUpperLoopAdapter(TrxChanThParams *params);
 
 /** Main drive threads */
-void *RxLowerLoopAdapter(Transceiver *);
-void *TxLowerLoopAdapter(Transceiver *);
+void *RxLowerLoopAdapter(Transceiver *transceiver);
+void *TxLowerLoopAdapter(Transceiver *transceiver);
 
 /** control message handler thread loop */
-void *ControlServiceLoopAdapter(TransceiverChannel *);
+void *ControlServiceLoopAdapter(TrxChanThParams *params);
 
 /** transmit queueing thread loop */
-void *TxUpperLoopAdapter(TransceiverChannel *);
+void *TxUpperLoopAdapter(TrxChanThParams *params);
