@@ -1,7 +1,10 @@
 /*
 * Copyright 2018 sysmocom - s.f.m.c. GmbH
 *
-* This software is distributed under multiple licenses; see the COPYING file in the main directory for licensing information for this specific distribuion.
+* SPDX-License-Identifier: AGPL-3.0+
+*
+* This software is distributed under multiple licenses; see the COPYING file in
+* the main directory for licensing information for this specific distribution.
 *
 * This use of this software may be subject to additional restrictions.
 * See the LEGAL file in the main directory for details.
@@ -42,18 +45,10 @@
 class LMSDevice:public RadioDevice {
 
 private:
-
-	static constexpr double masterClockRate = 52.0e6;
-
 	lms_device_t *m_lms_dev;
 	std::vector<lms_stream_t> m_lms_stream_rx;
 	std::vector<lms_stream_t> m_lms_stream_tx;
 
-	std::vector<uint32_t> m_last_rx_underruns;
-	std::vector<uint32_t> m_last_rx_overruns;
-	std::vector<uint32_t> m_last_rx_dropped;
-	std::vector<uint32_t> m_last_tx_underruns;
-
 	std::vector<smpl_buf *> rx_buffers;
 
 	double actualSampleRate;	///< the actual USRP sampling rate
@@ -63,14 +58,16 @@ private:
 
 	TIMESTAMP ts_initial, ts_offset;
 
-	double rxGain;
+	std::vector<double> tx_gains, rx_gains;
 	double maxTxGainClamp;
 
 	bool do_calib(size_t chan);
 	bool do_filters(size_t chan);
+	void log_ant_list(bool dir_tx, size_t chan, std::ostringstream& os);
 	int get_ant_idx(const std::string & name, bool dir_tx, size_t chan);
 	bool flush_recv(size_t num_pkts);
-	void update_stream_stats(size_t chan, bool * underrun, bool * overrun);
+	void update_stream_stats_rx(size_t chan, bool *overrun);
+	void update_stream_stats_tx(size_t chan, bool *underrun);
 
 public:
 
@@ -89,10 +86,6 @@ public:
 	/** Stop the LMS */
 	bool stop();
 
-	/** Set priority not supported */
-	void setPriority(float prio = 0.5) {
-	}
-
 	enum TxWindowType getWindowType() {
 		return TX_WINDOW_LMS1;
 	}
@@ -157,7 +150,7 @@ public:
 
 	/** get the current receive gain */
 	double getRxGain(size_t chan = 0) {
-		return rxGain;
+		return rx_gains[chan];
 	}
 
 	/** return maximum Rx Gain **/
@@ -169,6 +162,11 @@ public:
 	/** sets the transmit chan gain, returns the gain setting **/
 	double setTxGain(double dB, size_t chan = 0);
 
+	/** get transmit gain */
+	double getTxGain(size_t chan = 0) {
+		return tx_gains[chan];
+	}
+
 	/** return maximum Tx Gain **/
 	double maxTxGain(void);
 
