@@ -6,6 +6,8 @@
  *
  * Author: Tom Tsou <tom.tsou@ettus.com>
  *
+ * SPDX-License-Identifier: AGPL-3.0+
+ *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
@@ -123,7 +125,6 @@ static const std::map<dev_key, dev_desc> dev_param_map {
 void *async_event_loop(uhd_device *dev)
 {
 	set_selfthread_name("UHDAsyncEvent");
-	dev->setPriority(0.43);
 
 	while (1) {
 		dev->recv_async_msg();
@@ -246,9 +247,6 @@ void uhd_device::set_rates()
 
 double uhd_device::setTxGain(double db, size_t chan)
 {
-	if (iface == MULTI_ARFCN)
-		chan = 0;
-
 	if (chan >= tx_gains.size()) {
 		LOGC(DDEV, ALERT) << "Requested non-existent channel" << chan;
 		return 0.0f;
@@ -280,9 +278,6 @@ double uhd_device::setTxGain(double db, size_t chan)
 
 double uhd_device::setRxGain(double db, size_t chan)
 {
-	if (iface == MULTI_ARFCN)
-		chan = 0;
-
 	if (chan >= rx_gains.size()) {
 		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return 0.0f;
@@ -298,9 +293,6 @@ double uhd_device::setRxGain(double db, size_t chan)
 
 double uhd_device::getRxGain(size_t chan)
 {
-	if (iface == MULTI_ARFCN)
-		chan = 0;
-
 	if (chan >= rx_gains.size()) {
 		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return 0.0f;
@@ -309,6 +301,16 @@ double uhd_device::getRxGain(size_t chan)
 	return rx_gains[chan];
 }
 
+double uhd_device::getTxGain(size_t chan)
+{
+	if (chan >= tx_gains.size()) {
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
+		return 0.0f;
+	}
+
+	return tx_gains[chan];
+}
+
 /*
     Parse the UHD device tree and mboard name to find out what device we're
     dealing with. We need the window type so that the transceiver knows how to
@@ -380,7 +382,6 @@ void uhd_device::set_channels(bool swap)
 		if (dev_type != B200 && dev_type != B210)
 			throw std::invalid_argument("Device does not support MCBTS");
 		dev_type = B2XX_MCBTS;
-		chans = 1;
 	}
 
 	if (chans > dev_param_map.at(dev_key(dev_type, tx_sps, rx_sps)).channels)
@@ -641,12 +642,6 @@ bool uhd_device::stop()
 	return true;
 }
 
-void uhd_device::setPriority(float prio)
-{
-	uhd::set_thread_priority_safe(prio);
-	return;
-}
-
 int uhd_device::check_rx_md_err(uhd::rx_metadata_t &md, ssize_t num_smpls)
 {
 	if (!num_smpls) {
@@ -747,7 +742,7 @@ int uhd_device::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 		for (size_t i = 0; i < rx_buffers.size(); i++) {
 			rc = rx_buffers[i]->write((short *) &pkt_bufs[i].front(),
 						  num_smpls,
-						  metadata.time_spec.to_ticks(rx_rate));
+						  ts.to_ticks(rx_rate));
 
 			// Continue on local overrun, exit on other errors
 			if ((rc < 0)) {
