@@ -1,6 +1,8 @@
 /*
 * Copyright 2008, 2009 Free Software Foundation, Inc.
 *
+* SPDX-License-Identifier: AGPL-3.0+
+*
 * This software is distributed under the terms of the GNU Affero Public License.
 * See the COPYING file in the main directory for details.
 *
@@ -69,6 +71,7 @@ USRPDevice::USRPDevice(size_t tx_sps, size_t rx_sps, InterfaceType iface,
   decimRate = (unsigned int) round(masterClockRate/((GSMRATE) * (double) tx_sps));
   actualSampleRate = masterClockRate/decimRate;
   rxGain = 0;
+  txGain = 0;
 
   /*
    * Undetermined delay b/w ping response timestamp and true
@@ -277,10 +280,11 @@ double USRPDevice::setTxGain(double dB, size_t chan)
 
   if (!m_dbTx->set_gain(dB))
     LOGC(DDEV, ERR) << "Error setting TX gain";
-
+  else
+    txGain = dB;
   writeLock.unlock();
 
-  return dB;
+  return txGain;
 }
 
 
@@ -303,10 +307,11 @@ double USRPDevice::setRxGain(double dB, size_t chan)
 
   if (!m_dbRx->set_gain(dB))
     LOGC(DDEV, ERR) << "Error setting RX gain";
-
+  else
+    rxGain = dB;
   writeLock.unlock();
 
-  return dB;
+  return rxGain;
 }
 
 bool USRPDevice::setRxAntenna(const std::string &ant, size_t chan)
@@ -375,7 +380,7 @@ int USRPDevice::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
     return len;
   }
 
-  if (underrun) *underrun = false;
+  *underrun = false;
 
   uint32_t readBuf[2000];
 
@@ -425,7 +430,7 @@ int USRPDevice::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 	continue;
       }
       if ((word0 >> 28) & 0x04) {
-	if (underrun) *underrun = true;
+	*underrun = true;
 	LOGC(DDEV, DEBUG) << "UNDERRUN in TRX->USRP interface";
       }
       if (RSSI) *RSSI = (word0 >> 21) & 0x3f;
