@@ -1,7 +1,10 @@
 /*
 * Copyright 2008 Free Software Foundation, Inc.
 *
-* This software is distributed under multiple licenses; see the COPYING file in the main directory for licensing information for this specific distribuion.
+* SPDX-License-Identifier: AGPL-3.0+
+*
+* This software is distributed under multiple licenses; see the COPYING file in
+* the main directory for licensing information for this specific distribution.
 *
 * This use of this software may be subject to additional restrictions.
 * See the LEGAL file in the main directory for details.
@@ -77,6 +80,7 @@ private:
   unsigned long lastPktTimestamp;
 
   double rxGain;
+  double txGain;
 
 #ifdef SWLOOPBACK
   short loopbackBuffer[1000000];
@@ -104,9 +108,6 @@ private:
   /** Stop the USRP */
   bool stop();
 
-  /** Set priority not supported */
-  void setPriority(float prio = 0.5) { }
-
   enum TxWindowType getWindowType() { return TX_WINDOW_USRP1; }
 
   /**
@@ -170,6 +171,9 @@ private:
   /** sets the transmit chan gain, returns the gain setting **/
   double setTxGain(double dB, size_t chan = 0);
 
+  /** get transmit gain */
+  double getTxGain(size_t chan = 0) { return txGain; }
+
   /** return maximum Tx Gain **/
   double maxTxGain(void);
 
