@@ -1,6 +1,8 @@
 /*
 * Copyright 2008, 2011 Free Software Foundation, Inc.
 *
+* SPDX-License-Identifier: AGPL-3.0+
+*
 * This software is distributed under the terms of the GNU Affero Public License.
 * See the COPYING file in the main directory for details.
 *
@@ -84,7 +86,7 @@ static Resampler *dnsampler = NULL;
  * perform 16-byte memory alignment required by many SSE instructions.
  */
 struct CorrelationSequence {
-  CorrelationSequence() : sequence(NULL)
+  CorrelationSequence() : sequence(NULL), buffer(NULL), toa(0.0)
   {
   }
 
@@ -343,7 +345,7 @@ static signalVector *convolve(const signalVector *x, const signalVector *h,
     _x = x;
 
   /*
-   * Four convovle types:
+   * Four convolve types:
    *   1. Complex-Real (aligned)
    *   2. Complex-Complex (aligned)
    *   3. Complex-Real (!aligned)
@@ -530,19 +532,17 @@ static PulseSequence *generateGSMPulse(int sps)
   return pulse;
 }
 
-bool vectorSlicer(SoftVector *x)
+/* Convert -1..+1 soft bits to 0..1 soft bits */
+void vectorSlicer(float *dest, const float *src, size_t len)
 {
-  SoftVector::iterator xP = x->begin();
-  SoftVector::iterator xPEnd = x->end();
-  while (xP < xPEnd) {
-    *xP = 0.5 * (*xP + 1.0f);
-    if (*xP > 1.0)
-      *xP = 1.0;
-    if (*xP < 0.0)
-      *xP = 0.0;
-    xP++;
-  }
-  return true;
+	size_t i;
+	for (i = 0; i < len; i++) {
+		dest[i] = 0.5 * (src[i] + 1.0f);
+		if (dest[i] > 1.0)
+			dest[i] = 1.0;
+		else if (dest[i] < 0.0)
+			dest[i] = 0.0;
+	}
 }
 
 static signalVector *rotateBurst(const BitVector &wBurst,
@@ -723,7 +723,7 @@ static signalVector *mapEdgeSymbols(const BitVector &bits)
  *
  * Delay the EDGE downlink bursts by one symbol in order to match GMSK pulse
  * shaping group delay. The difference in group delay arises from the dual
- * pulse filter combination of the GMSK Laurent represenation whereas 8-PSK
+ * pulse filter combination of the GMSK Laurent representation whereas 8-PSK
  * uses a single pulse linear filter.
  */
 static signalVector *shapeEdgeBurst(const signalVector &symbols)
@@ -1455,6 +1455,33 @@ static signalVector *downsampleBurst(const signalVector &burst)
   return out;
 };
 
+/*
+ * Computes C/I (Carrier-to-Interference ratio) in dB (deciBels).
+ * It is computed from the training sequence of each received burst,
+ * by comparing the "ideal" training sequence with the actual one.
+ */
+static float computeCI(const signalVector *burst, CorrelationSequence *sync,
+                       float toa, int start, complex xcorr)
+{
+  float S, C;
+  int ps;
+
+  /* Integer position where the sequence starts */
+  ps = start + 1 - sync->sequence->size() + (int)roundf(toa);
+
+  /* Estimate Signal power */
+  S = 0.0f;
+  for (int i=0, j=ps; i<(int)sync->sequence->size(); i++,j++)
+    S += (*burst)[j].norm2();
+  S /= sync->sequence->size();
+
+  /* Esimate Carrier power */
+  C = xcorr.norm2() / ((sync->sequence->size() - 1) * sync->gain.abs());
+
+  /* Interference = Signal - Carrier, so C/I = C / (S - C) */
+  return 3.0103f * log2f(C / (S - C));
+}
+
 /*
  * Detect a burst based on correlation and peak-to-average ratio
  *
@@ -1465,11 +1492,13 @@ static signalVector *downsampleBurst(const signalVector &burst)
  */
 static int detectBurst(const signalVector &burst,
                        signalVector &corr, CorrelationSequence *sync,
-                       float thresh, int sps, complex *amp, float *toa,
-                       int start, int len)
+                       float thresh, int sps, int start, int len,
+                       struct estim_burst_params *ebp)
 {
   const signalVector *corr_in;
   signalVector *dec = NULL;
+  complex xcorr;
+  int rc = 1;
 
   if (sps == 4) {
     dec = downsampleBurst(burst);
@@ -1482,35 +1511,42 @@ static int detectBurst(const signalVector &burst,
   /* Correlate */
   if (!convolve(corr_in, sync->sequence, &corr,
                 CUSTOM, start, len)) {
-    delete dec;
-    return -1;
+    rc = -1;
+    goto del_ret;
   }
 
-  delete dec;
-
   /* Running at the downsampled rate at this point */
   sps = 1;
 
   /* Peak detection - place restrictions at correlation edges */
-  *amp = fastPeakDetect(corr, toa);
+  ebp->amp = fastPeakDetect(corr, &ebp->toa);
 
-  if ((*toa < 3 * sps) || (*toa > len - 3 * sps))
-    return 0;
+  if ((ebp->toa < 3 * sps) || (ebp->toa > len - 3 * sps)) {
+    rc = 0;
+    goto del_ret;
+  }
 
-  /* Peak -to-average ratio */
-  if (computePeakRatio(&corr, sps, *toa, *amp) < thresh)
-    return 0;
+  /* Peak-to-average ratio */
+  if (computePeakRatio(&corr, sps, ebp->toa, ebp->amp) < thresh) {
+    rc = 0;
+    goto del_ret;
+  }
 
-  /* Compute peak-to-average ratio. Reject if we don't have enough values */
-  *amp = peakDetect(corr, toa, NULL);
+  /* Refine TOA and correlation value */
+  xcorr = peakDetect(corr, &ebp->toa, NULL);
+
+  /* Compute C/I */
+  ebp->ci = computeCI(corr_in, sync, ebp->toa, start, xcorr);
 
   /* Normalize our channel gain */
-  *amp = *amp / sync->gain;
+  ebp->amp = xcorr / sync->gain;
 
   /* Compensate for residuate time lag */
-  *toa = *toa - sync->toa;
+  ebp->toa = ebp->toa - sync->toa;
 
-  return 1;
+del_ret:
+  delete dec;
+  return rc;
 }
 
 static float maxAmplitude(const signalVector &burst)
@@ -1534,13 +1570,10 @@ static float maxAmplitude(const signalVector &burst)
  *   head: Search symbols before target
  *   tail: Search symbols after target
  */
-static int detectGeneralBurst(const signalVector &rxBurst,
-                              float thresh,
-                              int sps,
-                              complex &amp,
-                              float &toa,
+static int detectGeneralBurst(const signalVector &rxBurst, float thresh, int sps,
                               int target, int head, int tail,
-                              CorrelationSequence *sync)
+                              CorrelationSequence *sync,
+                              struct estim_burst_params *ebp)
 {
   int rc, start, len;
   bool clipping = false;
@@ -1562,17 +1595,18 @@ static int detectGeneralBurst(const signalVector &rxBurst,
   signalVector corr(len);
 
   rc = detectBurst(rxBurst, corr, sync,
-                   thresh, sps, &amp, &toa, start, len);
+                   thresh, sps, start, len, ebp);
   if (rc < 0) {
     return -SIGERR_INTERNAL;
   } else if (!rc) {
-    amp = 0.0f;
-    toa = 0.0f;
+    ebp->amp = 0.0f;
+    ebp->toa = 0.0f;
+    ebp->ci = 0.0f;
     return clipping?-SIGERR_CLIP:SIGERR_NONE;
   }
 
   /* Subtract forward search bits from delay */
-  toa -= head;
+  ebp->toa -= head;
 
   return 1;
 }
@@ -1587,7 +1621,7 @@ static int detectGeneralBurst(const signalVector &rxBurst,
  *   tail: Search 8 symbols + maximum expected delay
  */
 static int detectRACHBurst(const signalVector &burst, float threshold, int sps,
-                           complex &amplitude, float &toa, unsigned max_toa, bool ext)
+                           unsigned max_toa, bool ext, struct estim_burst_params *ebp)
 {
   int rc, target, head, tail;
   int i, num_seq;
@@ -1598,10 +1632,12 @@ static int detectRACHBurst(const signalVector &burst, float threshold, int sps,
   num_seq = ext ? 3 : 1;
 
   for (i = 0; i < num_seq; i++) {
-    rc = detectGeneralBurst(burst, threshold, sps, amplitude, toa,
-                            target, head, tail, gRACHSequences[i]);
-    if (rc > 0)
+    rc = detectGeneralBurst(burst, threshold, sps, target, head, tail,
+                            gRACHSequences[i], ebp);
+    if (rc > 0) {
+      ebp->tsc = i;
       break;
+    }
   }
 
   return rc;
@@ -1616,7 +1652,7 @@ static int detectRACHBurst(const signalVector &burst, float threshold, int sps,
  *   tail: Search 6 symbols + maximum expected delay
  */
 static int analyzeTrafficBurst(const signalVector &burst, unsigned tsc, float threshold,
-                               int sps, complex &amplitude, float &toa, unsigned max_toa)
+                               int sps, unsigned max_toa, struct estim_burst_params *ebp)
 {
   int rc, target, head, tail;
   CorrelationSequence *sync;
@@ -1629,13 +1665,13 @@ static int analyzeTrafficBurst(const signalVector &burst, unsigned tsc, float th
   tail = 6 + max_toa;
   sync = gMidambles[tsc];
 
-  rc = detectGeneralBurst(burst, threshold, sps, amplitude, toa,
-                          target, head, tail, sync);
+  ebp->tsc = tsc;
+  rc = detectGeneralBurst(burst, threshold, sps, target, head, tail, sync, ebp);
   return rc;
 }
 
 static int detectEdgeBurst(const signalVector &burst, unsigned tsc, float threshold,
-                           int sps, complex &amplitude, float &toa, unsigned max_toa)
+                           int sps, unsigned max_toa, struct estim_burst_params *ebp)
 {
   int rc, target, head, tail;
   CorrelationSequence *sync;
@@ -1648,33 +1684,31 @@ static int detectEdgeBurst(const signalVector &burst, unsigned tsc, float thresh
   tail = 6 + max_toa;
   sync = gEdgeMidambles[tsc];
 
-  rc = detectGeneralBurst(burst, threshold, sps, amplitude, toa,
-                          target, head, tail, sync);
+  ebp->tsc = tsc;
+  rc = detectGeneralBurst(burst, threshold, sps,
+                          target, head, tail, sync, ebp);
   return rc;
 }
 
 int detectAnyBurst(const signalVector &burst, unsigned tsc, float threshold,
-                   int sps, CorrType type, complex &amp, float &toa,
-                   unsigned max_toa)
+                   int sps, CorrType type, unsigned max_toa,
+                   struct estim_burst_params *ebp)
 {
   int rc = 0;
 
   switch (type) {
   case EDGE:
-    rc = detectEdgeBurst(burst, tsc, threshold, sps,
-                         amp, toa, max_toa);
+    rc = detectEdgeBurst(burst, tsc, threshold, sps, max_toa, ebp);
     if (rc > 0)
       break;
     else
       type = TSC;
   case TSC:
-    rc = analyzeTrafficBurst(burst, tsc, threshold, sps,
-                             amp, toa, max_toa);
+    rc = analyzeTrafficBurst(burst, tsc, threshold, sps, max_toa, ebp);
     break;
   case EXT_RACH:
   case RACH:
-    rc = detectRACHBurst(burst, threshold, sps, amp, toa,
-                         max_toa, type == EXT_RACH);
+    rc = detectRACHBurst(burst, threshold, sps, max_toa, type == EXT_RACH, ebp);
     break;
   default:
     LOG(ERR) << "Invalid correlation type";
